from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

from nuclia.exceptions import KBNotAvailable
from nuclia.lib.kb import AsyncNucliaDBClient, Environment, NucliaDBClient

if TYPE_CHECKING:
    from nuclia.config import Config
    from nuclia.sdk.auth import AsyncNucliaAuth, NucliaAuth


@dataclass
class DataConfig:
    config: Optional[Config] = None
    auth: Optional[NucliaAuth] = None
    async_auth: Optional[AsyncNucliaAuth] = None


DATA = DataConfig()


def set_config(config: Config):
    DATA.config = config


def get_config() -> Config:
    if DATA.config is None:
        from nuclia.config import read_config

        DATA.config = read_config()
    return DATA.config


def get_auth() -> NucliaAuth:
    get_config()
    if DATA.auth is None:
        from nuclia.sdk.auth import NucliaAuth

        DATA.auth = NucliaAuth()
    return DATA.auth


def get_async_auth() -> AsyncNucliaAuth:
    get_config()
    if DATA.async_auth is None:
        from nuclia.sdk.auth import AsyncNucliaAuth

        DATA.async_auth = AsyncNucliaAuth()
    return DATA.async_auth


def get_client(kbid: str) -> NucliaDBClient:
    auth = get_auth()
    kb_obj = auth._config.get_kb(kbid)

    if kb_obj is None:
        raise KBNotAvailable(kbid)
    elif kb_obj.region is None:
        # OSS
        ndb = NucliaDBClient(environment=Environment.OSS, url=kb_obj.url)
    else:
        if kb_obj.token is None and auth._validate_user_token():
            # User token auth
            ndb = NucliaDBClient(
                environment=Environment.CLOUD,
                url=kb_obj.url,
                user_token=auth._config.token,
                region=kb_obj.region,
            )
        elif kb_obj.token is None:
            # Public
            ndb = NucliaDBClient(
                environment=Environment.CLOUD,
                url=kb_obj.url,
                region=kb_obj.region,
            )
        else:
            ndb = NucliaDBClient(
                environment=Environment.CLOUD,
                url=kb_obj.url,
                api_key=kb_obj.token,
                region=kb_obj.region,
            )
    return ndb


def get_async_client(kbid: str) -> AsyncNucliaDBClient:
    auth = get_async_auth()
    kb_obj = auth._config.get_kb(kbid)

    if kb_obj is None:
        raise KBNotAvailable(kbid)
    elif kb_obj.region is None:
        # OSS
        ndb = AsyncNucliaDBClient(environment=Environment.OSS, url=kb_obj.url)
    else:
        if kb_obj.token is None and auth._validate_user_token():
            # User token auth
            ndb = AsyncNucliaDBClient(
                environment=Environment.CLOUD,
                url=kb_obj.url,
                user_token=auth._config.token,
                region=kb_obj.region,
            )
        elif kb_obj.token is None:
            # Public
            ndb = AsyncNucliaDBClient(
                environment=Environment.CLOUD,
                url=kb_obj.url,
                region=kb_obj.region,
            )
        else:
            ndb = AsyncNucliaDBClient(
                environment=Environment.CLOUD,
                url=kb_obj.url,
                api_key=kb_obj.token,
                region=kb_obj.region,
            )
    return ndb



import logging
import os


def yes_no(question: str) -> bool:
    if os.environ.get("TESTING") == "True":
        return True
    yes_choices = ["yes", "y"]
    no_choices = ["no", "n"]
    value = None

    while True:
        user_input = input(f"{question} (yes/no): ")
        if user_input.lower() in yes_choices:
            value = True
            break
        elif user_input.lower() in no_choices:
            value = False
            break
        else:
            print("Type yes or no")
            continue
    return value


class CustomFormatter(logging.Formatter):
    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format_template = "%(message)s"

    FORMATS = {
        logging.DEBUG: grey + format_template + reset,
        logging.INFO: grey + format_template + reset,
        logging.WARNING: yellow + format_template + reset,
        logging.ERROR: red + format_template + reset,
        logging.CRITICAL: bold_red + format_template + reset,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)



import json
from typing import Union

import httpx
import requests
from tabulate import tabulate

from nuclia.exceptions import RateLimitError, UserTokenExpired, DuplicateError
from nucliadb_models.resource import ResourceList
from nucliadb_models.search import SyncAskResponse


def handle_http_errors(response: Union[httpx.Response, requests.models.Response]):
    if (
        response.status_code == 403
        and "Hydra token is either unexistent or revoked" in response.text
    ):
        raise UserTokenExpired()
    elif response.status_code == 429:
        raise RateLimitError(f"Rate limited: {response.text}")
    elif response.status_code == 409:
        raise DuplicateError("Duplicate resource")
    elif response.status_code >= 400:
        raise httpx.HTTPError(f"Status code {response.status_code}: {response.text}")


def serialize(obj):
    if isinstance(obj, ResourceList):
        data = []
        for resource in obj.resources:
            status = resource.metadata.status if resource.metadata is not None else ""
            data.append(
                [resource.id, resource.icon, resource.title, status, resource.slug]
            )
        return tabulate(
            data,
            headers=["UUID", "Icon", "Title", "Status", "Slug"],
        )
    if isinstance(obj, SyncAskResponse):
        if obj.status != "success":
            response = f"ERROR: {obj.error_details}"
        elif obj.answer_json is not None:
            response = f"JSON: {json.dumps(obj.answer_json, indent=2)} \n"
        else:
            response = f"Answer: {obj.answer} \n"

        if obj.metadata:
            response += f"Metadata: {obj.metadata.model_dump_json(indent=2)}\n"

        response += "Resources: \n"
        data = []
        for resource in obj.retrieval_results.resources.values():
            data.append([resource.id, resource.icon, resource.title])
        response += tabulate(
            data,
            headers=["UUID", "Icon", "Title"],
        )
        return response
    return obj



import tempfile
import time

import pytest

from nuclia.sdk.export_import import NucliaExports, NucliaImports


@pytest.mark.skip(reason="To avoid duplicating data in the test KB")
def test_sync(testing_config):
    exports = NucliaExports()
    imports = NucliaImports()

    with tempfile.TemporaryDirectory() as tempdir:
        path = f"{tempdir}/kb.export"
        exports.start(path=path)
        imports.start(path=path, sync=True)


@pytest.mark.skip(reason="To avoid duplicating data in the test KB")
def test_manual(testing_config):
    exports = NucliaExports()
    imports = NucliaImports()
    with tempfile.TemporaryDirectory() as tempdir:
        path = f"{tempdir}/kb.export"
        resp = exports.start(path=path)
        exports.download(resp.export_id, path=path)
        imports.start(path=path)
        while True:
            status = imports.status(resp.import_id).status
            if status == "finished":
                break
            assert status != "failed"
            time.sleep(5)



import asyncio
from functools import wraps
import inspect

from httpx import ConnectError
import yaml

from nuclia import BASE_DOMAIN
from nuclia.data import get_async_auth, get_async_client, get_auth, get_client
from nuclia.exceptions import NotDefinedDefault, NucliaConnectionError
from nuclia.lib.kb import AsyncNucliaDBClient, Environment, NucliaDBClient
from nuclia.lib.nua import AsyncNuaClient, NuaClient


def accounts(func):
    @wraps(func)
    def wrapper_checkout_accounts(*args, **kwargs):
        auth = get_auth()
        auth.accounts()
        return func(*args, **kwargs)

    return wrapper_checkout_accounts


def kbs(func):
    @wraps(func)
    def wrapper_checkout_kbs(*args, **kwargs):
        if "account" in kwargs:
            auth = get_auth()
            account_id = auth.get_account_id(kwargs["account"])
            auth.kbs(account_id)
        return func(*args, **kwargs)

    return wrapper_checkout_kbs


def kb(func):
    @wraps(func)
    async def async_wrapper_checkout_kb(*args, **kwargs):
        if "ndb" in kwargs:
            return await func(*args, **kwargs)
        url = kwargs.get("url")
        api_key = kwargs.get("api_key")
        auth = get_async_auth()
        if url is None:
            # Get default KB
            kbid = auth._config.get_default_kb()
            if kbid is None:
                raise NotDefinedDefault()
            ndb = get_async_client(kbid)
        elif url.find(BASE_DOMAIN) >= 0:
            region = url.split(".")[0].split("/")[-1]
            ndb = AsyncNucliaDBClient(
                environment=Environment.CLOUD, url=url, api_key=api_key, region=region
            )
        else:
            ndb = AsyncNucliaDBClient(environment=Environment.OSS, url=url)
        kwargs["ndb"] = ndb
        try:
            result = await func(*args, **kwargs)
            return result
        except ConnectError:
            raise NucliaConnectionError(f"Could not connect to {ndb}")

    @wraps(func)
    def wrapper_checkout_kb(*args, **kwargs):
        if "ndb" in kwargs:
            return func(*args, **kwargs)
        url = kwargs.get("url")
        api_key = kwargs.get("api_key")
        auth = get_auth()
        if url is None:
            # Get default KB
            kbid = auth._config.get_default_kb()
            if kbid is None:
                raise NotDefinedDefault()
            ndb = get_client(kbid)
        elif url.find(BASE_DOMAIN) >= 0:
            region = url.split(".")[0].split("/")[-1]
            ndb = NucliaDBClient(
                environment=Environment.CLOUD, url=url, api_key=api_key, region=region
            )
        else:
            ndb = NucliaDBClient(environment=Environment.OSS, url=url)
        kwargs["ndb"] = ndb
        try:
            result = func(*args, **kwargs)
            return result
        except ConnectError:
            raise NucliaConnectionError(f"Could not connect to {ndb}")

    if asyncio.iscoroutinefunction(func):
        return async_wrapper_checkout_kb
    else:
        return wrapper_checkout_kb


def nucliadb(func):
    @wraps(func)
    def wrapper_checkout_nucliadb(*args, **kwargs):
        if "ndb" in kwargs:
            return func(*args, **kwargs)
        url = kwargs.get("url")
        auth = get_auth()
        if url is None:
            # Get default KB
            nucliadb = auth._config.get_default_nucliadb()
            if nucliadb is None:
                raise NotDefinedDefault()
            # OSS
            ndb = NucliaDBClient(
                environment=Environment.OSS, base_url=f"{nucliadb}/api"
            )
        else:
            ndb = NucliaDBClient(environment=Environment.OSS, base_url=f"{url}/api")
        kwargs["ndb"] = ndb
        return func(*args, **kwargs)

    return wrapper_checkout_nucliadb


def nua(func):
    @wraps(func)
    async def async_wrapper_checkout_nua(*args, **kwargs):
        auth = get_auth()
        nua_id = auth._config.get_default_nua()
        if nua_id is None:
            raise NotDefinedDefault()

        nua_obj = auth._config.get_nua(nua_id)
        nc = AsyncNuaClient(
            region=nua_obj.region, account=nua_obj.account, token=nua_obj.token
        )

        kwargs["nc"] = nc
        return await func(*args, **kwargs)

    @wraps(func)
    async def async_generative_wrapper_checkout_nua(*args, **kwargs):
        auth = get_auth()
        nua_id = auth._config.get_default_nua()
        if nua_id is None:
            raise NotDefinedDefault()

        nua_obj = auth._config.get_nua(nua_id)
        nc = AsyncNuaClient(
            region=nua_obj.region, account=nua_obj.account, token=nua_obj.token
        )

        kwargs["nc"] = nc
        async for value in func(*args, **kwargs):
            yield value

    @wraps(func)
    def wrapper_checkout_nua(*args, **kwargs):
        auth = get_auth()
        nua_id = auth._config.get_default_nua()
        if nua_id is None:
            raise NotDefinedDefault()

        nua_obj = auth._config.get_nua(nua_id)
        nc = NuaClient(
            region=nua_obj.region, account=nua_obj.account, token=nua_obj.token
        )

        kwargs["nc"] = nc
        return func(*args, **kwargs)

    if inspect.isasyncgenfunction(func):
        return async_generative_wrapper_checkout_nua
    elif asyncio.iscoroutinefunction(func):
        return async_wrapper_checkout_nua
    else:
        return wrapper_checkout_nua


def account(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        account_slug = kwargs.get("account")
        account_id = kwargs.get("account_id")
        auth = get_auth()
        if account_id is None and account_slug is None:
            account_slug = auth._config.get_default_account()
            if account_slug is None:
                raise NotDefinedDefault()
            kwargs["account"] = account_slug
        if account_id is None:
            account_id = auth.get_account_id(account_slug)  # type: ignore
            kwargs["account_id"] = account_id
        return func(*args, **kwargs)

    return wrapper


def pretty(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if kwargs.get("json"):
            return result.json(indent=2)
        if kwargs.get("yaml"):
            return yaml.dump(result)
        return result

    @wraps(func)
    async def async_wrapper(*args, **kwargs):
        result = await func(*args, **kwargs)
        if kwargs.get("json"):
            return result.json(indent=2)
        if kwargs.get("yaml"):
            return yaml.dump(result)
        return result

    if asyncio.iscoroutinefunction(func):
        return async_wrapper
    else:
        return wrapper


def zone(func):
    @wraps(func)
    def wrapper_checkout_zone(*args, **kwargs):
        zone = kwargs.get("zone")
        if not zone:
            auth = get_auth()
            kwargs["zone"] = auth._config.get_default_zone()
        return func(*args, **kwargs)

    return wrapper_checkout_zone



from nuclia.decorators import kb
from nuclia.lib.kb import LogType, NucliaDBClient


class NucliaLogs:
    @kb
    def get(self, *args, type: LogType, month: str, **kwargs) -> list[list[str]]:
        """
        Download activity logs.

        :param type: NEW, PROCESSED, MODIFIED, CHAT, SEARCH, FEEDBACK
        :param month: YYYY-MM
        """
        ndb: NucliaDBClient = kwargs["ndb"]
        resp = ndb.logs(type=type, month=month)
        return resp
